import numpy as np
import torch
import random
from neurodiffeq.utils import set_seed


def test_set_seed():
    set_seed(10)
    x1 = np.random.rand(5)
    y1 = torch.rand(5)
    z1 = random.random()

    set_seed(10)
    x2 = np.random.rand(5)
    y2 = torch.rand(5)
    z2 = random.random()

    assert (x1 == x2).all()
    assert (y1 == y2).all()
    assert z1 == z2



import torch
from torch import nn, optim
from neurodiffeq import diff
from neurodiffeq.networks import FCNN
from neurodiffeq.temporal import generator_3dspatial_body, generator_3dspatial_surface, generator_temporal
from neurodiffeq.temporal import FirstOrderInitialCondition, BoundaryCondition
from neurodiffeq.temporal import SingleNetworkApproximator3DSpatialTemporal
from neurodiffeq.temporal import MonitorMinimal
from neurodiffeq.temporal import _solve_3dspatial_temporal

def some_3d_time_dependent_pde(u, x, y, z, t):
    return diff(u, x) + diff(u, y) + diff(u, z) + diff(u, t) ...

# e.g. make u(x, y, z, t) = x^2 +y^2 + z^2 at the boundary
boundary_surface_1 = BoundaryCondition(
    form=lambda u, x, y, z: u - (x**2 + y**2 + z**2),
    points_generator=generator_3dspatial_surface( ... )
)
boundary_surface_2 = BoundaryCondition(
    form=lambda u, x, y, z: u - (x**2 + y**2 + z**2),
    points_generator=generator_3dspatial_surface( ... )
)
boundary_surface_3 = BoundaryCondition(
    form=lambda u, x, y, z: u - (x**2 + y**2 + z**2),
    points_generator=generator_3dspatial_surface( ... )
)

fcnn = FCNN(
    n_input_units=4,
    n_output_units=1,
    n_hidden_units=32,
    n_hidden_layers=1,
    actv=nn.Tanh
)
fcnn_approximator = SingleNetworkApproximator3DSpatialTemporal(
    single_network=fcnn,
    pde=some_3d_time_dependent_pde,
    boundary_conditions=[
        boundary_surface_1,
        boundary_surface_2,
        boundary_surface_3,
    ]
)
adam = optim.Adam(fcnn_approximator.parameters(), lr=0.001)

train_gen_spatial = generator_3dspatial_body(...)
train_gen_temporal = generator_temporal(...)
valid_gen_spatial = generator_3dspatial_body(...)
valid_gen_temporal = generator_temporal(...)

some_3d_time_dependent_pde_solution, _ = _solve_3dspatial_temporal(
    train_generator_spatial=train_gen_spatial,
    train_generator_temporal=train_gen_temporal,
    valid_generator_spatial=valid_gen_spatial,
    valid_generator_temporal=valid_gen_temporal,
    approximator=fcnn_approximator,
    optimizer=adam,
    batch_size=512,
    max_epochs=5000,
    shuffle=True,
    metrics={},
    monitor=MonitorMinimal(check_every=10)
)


# the model and the loss
# should be similar to `SingleNetworkApproximator2DSpatialTemporal`
class SingleNetworkApproximator3DSpatialTemporal(Approximator):
    def __init__(self, single_network, pde, initial_condition, boundary_conditions, boundary_strictness=1.):
        self.single_network = single_network
        self.pde = pde
        self.u0 = initial_condition.u0
        self.u0dot = initial_condition.u0dot if hasattr(initial_condition, 'u0dot') else None
        self.boundary_conditions = boundary_conditions
        self.boundary_strictness = boundary_strictness

    def __call__(self, xx, yy, zz, tt):
        xx = torch.unsqueeze(xx, dim=1)
        yy = torch.unsqueeze(yy, dim=1)
        zz = torch.unsqueeze(zz, dim=1)
        tt = torch.unsqueeze(tt, dim=1)
        xyzt = torch.cat((xx, yy, zz, tt), dim=1)
        uu_raw = self.single_network(xyzt)
        return uu_raw  # TODO reparameterize the output uu_raw (use the information carried by self.u0 and self.u0dot)

    def parameters(self):
        return self.single_network.parameters()

    def calculate_loss(self, xx, yy, zz, tt, x, y, z, t):
        uu = self.__call__(xx, yy, zz, tt)
        equation_mse = torch.mean(self.pde(uu, xx, yy, zz, tt)**2)
        return equation_mse  # TODO plus the regularization term (use the information carried by self.boundary_conditions)

    def calculate_metrics(self, xx, yy, zz, tt, metrics):
        uu = self.__call__(xx, yy, zz)

        return {
            metric_name: metric_func(*uu, xx, yy, zz)
            for metric_name, metric_func in metrics.items()
        }


# A generator for generating 3D points in the problem domain: yield (xx, yy, zz) where xx, yy, zz are 1-D tensors
def generator_3dspatial_body(...):
    pass

# A generator for generating 3D points on the boundary: yield (xx, yy, zz) where xx, yy, zz are 1-D tensors
def generator_3dspatial_surface(...):
    pass


def _solve_3dspatial_temporal(
    train_generator_spatial, train_generator_temporal, valid_generator_spatial, valid_generator_temporal,
    approximator, optimizer, batch_size, max_epochs, shuffle, metrics, monitor
):
    return _solve_spatial_temporal(
        train_generator_spatial, train_generator_temporal, valid_generator_spatial, valid_generator_temporal,
        approximator, optimizer, batch_size, max_epochs, shuffle, metrics, monitor,
        train_routine=_train_3dspatial_temporal, valid_routine=_valid_3dspatial_temporal
    )


# the logic for training one epoch
# should be similar to `_train_2dspatial_temporal`
def _train_3dspatial_temporal(train_generator_spatial, train_generator_temporal, approximator, optimizer, metrics, shuffle, batch_size):
    
    # generate x, y, z dimensions from train_generator_spatial
    # generate time slices from train_generator_temporal
    # Do a cartesian product of the above two to make the training set
    # pass the training points to the neural network in batches
    # for each batch calculate the loss and update the weights
    # calculate the loss custom metrics of this epoch

    return epoch_loss, epoch_metrics


# the logic for training one epoch
# should be similar to `_valid_2dspatial_temporal`
def _valid_3dspatial_temporal(valid_generator_spatial, valid_generator_temporal, approximator, metrics):

    # generate x, y, z dimensions from train_generator_spatial
    # generate time slices from train_generator_temporal
    # Do a cartesian product of the above two to make the training set
    # calculate the loss custom metrics of this epoch

    return epoch_loss, epoch_metrics
